/**
* @file  df_log.h
* @brief Schnittstelle fuer die Klasse CDFLog.
*/
//////////////////////////////////////////////////////////////////////
#pragma once
#ifndef DF_LOG_H
#define DF_LOG_H

#include <ctime>
#include <string>

// Initialisierungsdatei verwalten
#include "file/file_ini.h"
// Hilfsklasse zur Verwaltung alter Logdateien
#include "df_log_check.h"

#include <mutex>
#include <functional>
#include <list>
#include <string>

/* Die Logklasse ist fuer ein standardmaessiges logen von Funktionsaufrufen,
* so dass Fehler in der Verarbeitung auch noch nach einigen Tagen 
* nachfolzogen werden koennten.
* Sie ist ueber eine Initialisierungsdatei mit grundinformationen versorgbar.
*
*/
// YYYY_MM_DD_DFComDLL_channelID.log
#define LOG_FILENAME         "%s/%s/20%02i_%02i_%02i_DFComDLL_%i.log"
#define LOG_SEARCH_FILENAME  "%s/%s/20??_??_??_DFComDLL_*.log"
#define LOG_DELETE_FILENAME  "%s/%s/%s"

#if defined(_MSC_VER)
#define fullpath             _fullpath
#define mkdir                _mkdir
#define stat                 _stat
#define finddata_t           _finddata_t
#define findfirst            _findfirst
#define findnext             _findnext
#define findclose            _findclose
//#define snprintf	         _snprintf
#define access               _access
#endif

class CDFLog
{
    bool m_enabled;                    // Log aktiv?
    std::string m_filePath;            // Dateiname der Logdatei
    bool m_eraseFiles;                 // Alte Dateien loeschen
    bool m_elapse;                     // Ausfuehrungszeit angeben
    int m_writeList;                   // Listendaten mitloggen
public:
    int m_dockingPreferType;           // Bevorzugte Version der Dockingstation
    int m_traceMode;                   // Tracing aktivieren
    bool m_checkOpenAndCloseBySameThread; // Das Oeffnen und Schliessen des Kanals muss durch den selben Thread erfolgen.
    bool m_createXMLHelpFile{};
    std::string m_docPath;
    std::string m_dbPath;       // Pfad zur Datenbank, in der die Informationen bzgl. eines Firmwareupdates bereitgestellt werden
    uint32_t m_dbPathType{};    // Type der Pfadangabe

public:
    CDFLog( int connectionNumber, std::function<void( const std::string& )> changeTraceFilePath );
    virtual ~CDFLog();

    // Ist Log aktiv?
    inline bool logEnabled()
    {
        return m_enabled;
    }

    // Erzeugen einer Pfadangabe
    void MakeFilePath( char *pathBuffer, char *fileName, char *fileExt );

    // Einlesen der Initialisierungsdatei
    void logIniRead();
    // Pruefen auf alte Dateien.
    void logCheck();

    // Schreiben einer Logzeile, wenn Log aktiv.
    void logBeginPure( const char *functionName, const char *format, ... );
    void logBegin( const char *functionName, const char *format, ... );
    void logBegin( const char *functionName, int32_t deviceNumber, const char *format, ... );
    void logContinue( const char *format, ... );
    void logEnd( const char *format, ... ); // Mit Zeilenumbruch

    void logAddInfo( const char *format, ... );
    void logAddWarning( const char *format, ... );
    void logAddError( const char *format, ... );

    char *logGetBasePath( char *basePath )
    {
        std::sprintf( basePath, "%s/%s", m_logCheck.m_basePath, m_logCheck.m_subFolder );
        return basePath;
    }
    int logIsWriteList() { return m_enabled ? m_writeList : 0; }

    std::function<void (const std::string & )> callBackChangeTraceFilePath;
    std::string currentLogFilePath();
    std::string currentTraceFilePath();
    void checkCurrentFilePaths();

protected:
    std::string entryPrefix( const time_t &t );
    // Zugewiesene Verbindungsnummer
    int m_connectionNumber;
    // Zu verwendende Initialisierungsdatei
    FileIni m_logIni;
    // Start des Logeintrags
    std::clock_t m_clockBegin;
    // Alte Logdateien verwalten
    CDFLogCheck m_logCheck;
    // Bereichssperre fuer Dateizugriff.
    std::mutex m_mutexFile;
    // Aktueller Funktionsname zum Logeintrag
    std::string m_functionName;
    // Eventuelle Gertenummer des Logeintrags
    uint8_t m_deviceNumber;
    // Stringliste fr zustzliche Meldungen
    static const std::string m_logBeginFormatPure;
    static const std::string m_logBeginFormat;
    static const std::string m_logBeginFormatWithoutDeviceNumber;
    static const std::string m_logAddFormat;
    static const std::string m_logAddFormatWithoutDeviceNumber;
    std::list<std::string> m_subEntries;

    void logAdd( const char *addType, const char *format, va_list &va );
};

#endif

