#include "common/device_update.h"
#include "common/utils.h"
#include "file/file_device_zip.h"
#include "file/interface/hardware_info_items.h"

#include <stdio.h>
#include <algorithm>

#include <boost/format.hpp>

ModuleInfo::ModuleInfo( const uint32_t id, const uint8_t revision, const uint8_t assemblyIndex )
    : m_moduleId( id )
    , m_revision( revision )
    , m_assemblyIndex( assemblyIndex )
{
}

uint32_t ModuleInfo::moduleId()
{
    return m_moduleId;
}

uint8_t ModuleInfo::revision()
{
    return m_revision;
}

uint8_t ModuleInfo::assemblyIndex()
{
    return m_assemblyIndex;
}


std::vector<char> DeviceUpdate::createAssemblyIndexArray()
{
    std::vector<char> arr;
    arr.push_back( ' ' );
    for ( int i = 0; i < 26; ++i )
    {
        arr.push_back( char( 'a' + i ) );
    }
    return arr;
}

std::vector<char> DeviceUpdate::s_assemblyIndexArray = createAssemblyIndexArray();

DeviceUpdate::DeviceUpdate()
{
}

DeviceUpdate::DeviceUpdate( const std::string &pathToDB, DBPathType type )
    : m_pathToDB( pathToDB )
    , m_pathType( type )
{
}

DeviceUpdate::~DeviceUpdate()
{
    std::string errorMsg;
    close( errorMsg );
}

bool DeviceUpdate::updatePathToDB( const std::string &pathToDB, DBPathType type, std::string &errorMsg )
{
    std::string utf8Path = Utils::to_utf8( pathToDB );
    if ( m_pathToDB.compare( utf8Path ) == 0 )
    {
        return true;
    }

    m_pathToDB = utf8Path;
    m_pathType = type;

    // Reconnect der Datenbank, wenn diese bereits geffnet war.
    return isAvailable() ? open( errorMsg ) : true;
}

void DeviceUpdate::resetInternalData()
{
    m_minVersion = SoftwarebuildVersion();
    m_fwList.clear();
    m_dbBlacklist.clear();
}

DeviceUpdate::DBUnzipResult DeviceUpdate::unzipDB( FileDeviceZip &zipFile, const std::string filePath, std::string &errorMsg )
{
    std::string tmpDBPath;      // Pfad zur entpackten Datenbank aus dem Update-Archiv/-Ordner
    std::string expectedMD5;    // erwarteter md5 der im Archiv gespeicherten FW-Update-Datenbank

    // Prfe das Archiv, ob eine Datenbank enthalten sein sollte.
    if ( !zipFile.updateDBAvailableZip( filePath, expectedMD5, errorMsg ) )
    {
        return DBUnzipResult::Ignore;  // damit alte DFZ-Container ohne Update-Datenbank weiterhin untersttzt werden
    }

    // Entpacken der FW-Update-Datenbank
    if ( !zipFile.unzipUpdateDB( filePath, expectedMD5, tmpDBPath ) )
    {
        errorMsg = "Die Datenbak konnte nicht entpackt werden.";
        return DBUnzipResult::Failed;
    }

    if ( !updatePathToDB( tmpDBPath, DeviceUpdate::DBPathType::Local, errorMsg ) )
    {
        return DBUnzipResult::Failed;
    }
    
    return DBUnzipResult::Ok;
}

bool DeviceUpdate::open( std::string &errorMsg )
{
    if ( m_pathToDB.empty() )
    {
        return false;
    }

    errorMsg.clear();
    if ( m_db != nullptr )
    {
        close( errorMsg );
    }

    if ( sqlite3_open_v2( m_pathToDB.c_str(), &m_db, SQLITE_OPEN_READONLY | SQLITE_OPEN_URI, 0 ) > 0 )
    {
        errorMsg += sqlite3_errmsg( m_db );
        return false;
    }

    if ( !prepareStatements() )
    {
        errorMsg += "Preparation of SQL statements failed. [";
        errorMsg += sqlite3_errmsg( m_db );
        errorMsg = "]";
        return false;
    }
    return true;
}

void DeviceUpdate::close( std::string &errorMsg )
{
    if ( m_db != nullptr )
    {
        if ( sqlite3_finalize( m_stmtGetModuleMinVersion ) == SQLITE_OK )
        {
            m_stmtGetModuleMinVersion = nullptr;
        }
        else
        {
            errorMsg = sqlite3_errmsg( m_db );
        }

        if ( sqlite3_finalize( m_stmtGetModuleBlacklistEntries ) == SQLITE_OK )
        {
            m_stmtGetModuleBlacklistEntries = nullptr;
        }
        else
        {
            errorMsg = sqlite3_errmsg( m_db );
        }

        if ( sqlite3_close_v2( m_db ) == SQLITE_OK )
        {
            sqlite3_shutdown();
            m_db = nullptr;
        }
        else
        {
            errorMsg = sqlite3_errmsg( m_db );
        }
    }
}

bool DeviceUpdate::isAvailable()
{
    return (m_db != nullptr);
}

bool DeviceUpdate::prepareStatements()
{
    if ( m_db )
    {
        std::string sql;

        // get module min version
        sql = "SELECT printf( '%.2d.%.2d.%.2d.%.2d', f.master, f.major, f.minor, f.build ) as version, m.revision, m.assemblyIndex "
            "FROM MinFWMapping mfwm "
            "INNER JOIN Module m ON mfwm.refModuleId = m.id "
            "INNER JOIN Firmware f ON mfwm.refFirmwareId = f.id "
            "WHERE m.moduleId = :id "
            "AND mfwm.refModuleId <= ( "
            "  SELECT m.id "
            "  FROM Module m "
            "  WHERE m.moduleId = :id "
            "  AND m.revision = :revision "
            "  AND m.assemblyIndex = :assemblyIndex "
            ") "
            "UNION "
            "SELECT printf( '%.2d.%.2d.%.2d.%.2d', f.master, f.major, f.minor, f.build ) as version, m.revision, m.assemblyIndex "
            "FROM WhitelistFW wfw "
            "INNER JOIN Module m ON wfw.refModuleId = m.id "
            "INNER JOIN Firmware f ON wfw.refFirmwareId = f.id "
            "WHERE wfw.refModuleId = ( "
            "  SELECT m.id "
            "  FROM Module m "
            "  WHERE m.moduleId = :id "
            "  AND m.revision = :revision "
            "  AND m.assemblyIndex = :assemblyIndex "
            ") "
            "ORDER BY m.revision DESC, m.assemblyIndex DESC, version ASC "
            "LIMIT 1";

        auto rc = sqlite3_prepare_v2( m_db, sql.c_str(), -1, &m_stmtGetModuleMinVersion, nullptr );
        if ( rc != SQLITE_OK )
        {
            DF_ASSERT( false, sql.c_str() );
            return false;
        }

        // get blacklist entries of the module
        sql = "SELECT printf( '%.2d.%.2d.%.2d.%.2d', master, major, minor, build ) as version "
            "FROM Firmware WHERE id IN ( "
            "SELECT refFirmwareId FROM BlacklistFW "
            "WHERE refModuleId = (SELECT id FROM Module WHERE moduleId = :id AND revision = :revision AND assemblyIndex = :assemblyIndex) "
            "ORDER BY refFirmwareId "
            ") "
            "ORDER BY version";

        rc = sqlite3_prepare_v2( m_db, sql.c_str(), -1, &m_stmtGetModuleBlacklistEntries, nullptr );
        if ( rc != SQLITE_OK )
        {
            DF_ASSERT( false, sql.c_str() );
            return false;
        }

        return (m_stmtGetModuleMinVersion != nullptr && m_stmtGetModuleBlacklistEntries != nullptr);
    }
    return false;
}

const std::string &DeviceUpdate::dbPath() const
{
    return m_pathToDB;
}

bool DeviceUpdate::addFwVersion( const SoftwarebuildVersion &version, const std::string &fileName )
{
    bool result = true;
    for ( const auto &item : m_fwList )
    {
        result &= (item.first != version);
    }

    if ( result )
    {
        m_fwList[version] = fileName;
    }
    return result;
}

const std::string &DeviceUpdate::latestFwFile() const
{
    static std::string dummyString = "";
    
    if ( m_fwList.empty() )
    {
        return dummyString;
    }

    return m_fwList.rbegin()->second;
}

bool DeviceUpdate::detectMinFWVersion( ItemBoard &item, std::list<std::pair<std::string, std::string>> &moduleList )
{
    // Standarddatenbank ffnen (falls noch nicht geschehen)
    if ( !isAvailable() )
    {
        std::string errorMsg;
        if ( !open( errorMsg ) )
        {
            m_minVersion.reset();
            DF_ASSERT( false, errorMsg.c_str() );
            return false;
        }
    }

    // durch alle Module gehen...
    for ( auto &child : item.childs() )
    {
        if ( !detectMinFWVersion( child, moduleList ) )
        {
            m_minVersion.reset();
            return false;
        }
    }

    bool result{ false };
    HARDWARECONFIG &hwConfig = item.hc();
    sqlite3_bind_int( m_stmtGetModuleMinVersion, sqlite3_bind_parameter_index( m_stmtGetModuleMinVersion, ":id" ), hwConfig.id );
    sqlite3_bind_int( m_stmtGetModuleMinVersion, sqlite3_bind_parameter_index( m_stmtGetModuleMinVersion, ":revision" ), hwConfig.revision );
    sqlite3_bind_text( m_stmtGetModuleMinVersion, sqlite3_bind_parameter_index( m_stmtGetModuleMinVersion, ":assemblyIndex" ), (const char *)(&s_assemblyIndexArray[hwConfig.index]), 1, SQLITE_TRANSIENT );

    // Die kleinste Version steht an erster Stelle. Daher muss nur der erste Eintrag des Ergebnisses analysiert werden.
    if ( sqlite3_step( m_stmtGetModuleMinVersion ) == SQLITE_ROW )
    {
        std::string minVersion = Utils::from_utf8( reinterpret_cast<const char *>(sqlite3_column_text( m_stmtGetModuleMinVersion, 0 )) );
        updateMinFwVersion( minVersion );
        moduleList.push_back( make_pair(
            boost::str( boost::format( "id: %6d, revision: %d, assemblyIndex: %s" ) % hwConfig.id % (int)hwConfig.revision % s_assemblyIndexArray[hwConfig.index] ),
            minVersion ) );
        result = true;
    }
    else
    {
        moduleList.push_back( make_pair(
            boost::str( boost::format( "id: %6d, revision: %d, assemblyIndex: %s" ) % hwConfig.id % (int)hwConfig.revision % s_assemblyIndexArray[hwConfig.index] ),
            "" ) );
    }
    sqlite3_reset( m_stmtGetModuleMinVersion );
    return result;
}

void DeviceUpdate::updateMinFwVersion( const std::string &version )
{
    if ( !version.empty() && (m_minVersion < version || m_minVersion == SoftwarebuildVersion()) )
    {
        m_minVersion = SoftwarebuildVersion( version );
    }
}

const SoftwarebuildVersion &DeviceUpdate::minFwVersion() const
{
    return m_minVersion;
}

bool DeviceUpdate::collectBlacklistDefinition( ItemBoard &item )
{
    // Standarddatenbank ffnen (falls noch nicht geschehen)
    if ( !isAvailable() )
    {
        std::string errorMsg;
        if ( !open( errorMsg ) )
        {
            m_dbBlacklist.clear();
            DF_ASSERT( false, errorMsg.c_str() );
            return false;
        }
    }

    // durch alle Module gehen...
    for ( auto &child : item.childs() )
    {
        if ( !collectBlacklistDefinition( child ) )
        {
            m_dbBlacklist.clear();
            return false;
        }
    }

    if ( m_stmtGetModuleBlacklistEntries != nullptr )
    {
        HARDWARECONFIG &hwConfig = item.hc();
        sqlite3_bind_int( m_stmtGetModuleBlacklistEntries, sqlite3_bind_parameter_index( m_stmtGetModuleBlacklistEntries, ":id" ), hwConfig.id );
        sqlite3_bind_int( m_stmtGetModuleBlacklistEntries, sqlite3_bind_parameter_index( m_stmtGetModuleBlacklistEntries, ":revision" ), hwConfig.revision );
        sqlite3_bind_text( m_stmtGetModuleBlacklistEntries, sqlite3_bind_parameter_index( m_stmtGetModuleBlacklistEntries, ":assemblyIndex" ), (const char *)(&s_assemblyIndexArray[hwConfig.index]), 1, SQLITE_TRANSIENT );

        std::string version;
        while ( sqlite3_step( m_stmtGetModuleBlacklistEntries ) == SQLITE_ROW )
        {
            version = Utils::from_utf8( reinterpret_cast<const char *>(sqlite3_column_text( m_stmtGetModuleBlacklistEntries, 0 )) );
            if ( !isVersionBlacklisted( version ) )
            {
                m_dbBlacklist.push_back( version );
            }
        }
        sqlite3_reset( m_stmtGetModuleBlacklistEntries );
    }
    return true;
}

bool DeviceUpdate::isVersionBlacklisted( const std::string &version )
{
    return std::find( m_dbBlacklist.begin(), m_dbBlacklist.end(), version ) != m_dbBlacklist.end();
}

const std::list<std::string> &DeviceUpdate::blacklistVersions() const
{
    return m_dbBlacklist;
}

std::string DeviceUpdate::selectValue( sqlite3 *db, const char *sql )
{
    std::string value;
    if ( db != nullptr && sql != nullptr )
    {
        sqlite3_stmt *res;
        if ( sqlite3_prepare_v2( db, sql, -1, &res, 0 ) == SQLITE_OK )
        {
            if ( sqlite3_step( res ) == SQLITE_ROW )
            {
                value = Utils::from_utf8( reinterpret_cast<const char *>(sqlite3_column_text( res, 0 )) );
            }
            sqlite3_finalize( res );
        }
    }
    return value;
}
